/*
 * Test directory monitoring operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <thread>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



static file_t *open_object(char *filename, flags_t flags, bool have_uid, sys_uid_t uid)
{
  /* With a user, open a user-specific file opener. */

  if (have_uid)
  {
    l4_cap_idx_t opener = client_open_for_user((user_t) {uid, uid, 0022});

    if (l4_is_invalid_cap(opener))
    {
      printf("Could not obtain opener for file.\n");
      return NULL;
    }

    /* Invoke the open method to receive the reference. */

    return client_open_using(filename, flags, opener);
  }
  else
  {
    return client_open(filename, flags);
  }
}



/* Open files in the directory given by filename. */

static void open_files(const char *filename, bool have_uid, sys_uid_t uid)
{
  char *buffer = (char *) calloc(strlen(filename) + 10, sizeof(char));

  if (buffer == NULL)
  {
    printf("Could not allocate buffer.\n");
    return;
  }

  for (int i = 0; i < 10; i++)
  {
    sprintf(buffer, "%s/file%02d", filename, i);

    open_object(buffer, O_CREAT | O_RDWR, have_uid, uid);
    sleep(1);
  }
}

/* Monitor files in the given directory. */

static void monitor_files(file_t *directory)
{
  long err = client_subscribe(directory, NOTIFY_FILE_OPENED, client_notifier_task());

  if (err)
  {
    printf("Could not subscribe to events on directory.\n");
    return;
  }

  while (1)
  {
    /* Wait for notification of content. */

    err = client_wait_file(directory, client_notifier_task());

    if (err)
    {
      printf("Error waiting for notifications: %s\n", l4sys_errtostr(err));
      return;
    }

    printf("Notified with conditions:%s\n", directory->notifications & NOTIFY_FILE_OPENED ? " file opened" : "");
  }
}



int main(int argc, char *argv[])
{
  if (argc < 2)
  {
    printf("Need a directory name and an optional user identifier (if used with a filesystem).\n");
    return 1;
  }

  char *filename = argv[1];
  bool have_uid = (argc > 2) && strlen(argv[2]);
  sys_uid_t uid = have_uid ? atoi(argv[2]) : 0;

  printf("Opening %s...\n", filename);

  file_t *directory = open_object(filename, O_DIRECTORY, have_uid, uid);

  if (directory == NULL)
  {
    printf("Could not open directory.\n");
    return 1;
  }

  /* Schedule threads. */

  std::thread *activities[2];

  activities[0] = new std::thread(open_files, filename, have_uid, uid);
  activities[1] = new std::thread(monitor_files, directory);

  for (int i = 0; i < 2; i++)
    activities[i]->join();

  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
